{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE ViewPatterns #-}
{-# OPTIONS_GHC -fno-warn-overlapping-patterns #-}
module Engine.EngineM
( EngineM
, checkExperimentCache
, EngineEnv(..)
, EngineLog(..)
, EngineState
, mkInitialEngineState
, runEngineM
, lookupExperimentResult
, storeExperimentResult
, storeExperimentHValueAndId
, tellLog
, nextRandomR
) where

import Control.Monad.RWS.Strict hiding (liftIO)
import Data.Int
import Util.FCLabels hiding (get)
import Data.Maybe
import Data.Sequence (Seq)
import Engine.EngineHValue
import Engine.ExperimentCache (ExperimentCache, Strategy)
import Engine.ExperimentResult
import Exception
import Prelude hiding ((.), mod, id)
import Control.Category
import GHC
import HscTypes
import Linker (extendLinkEnv)
import MonadUtils
import Plan
import qualified Data.Sequence as Seq
import qualified Engine.ExperimentCache as EC
import System.Random
import UniqSupply
import Util.GHC()
import Var
import Debug.Trace

newtype EngineM a = EngineM { _runEngineM :: RWST EngineEnv EngineLog EngineState Ghc a }
  deriving (Monad, MonadReader EngineEnv, Functor) 

newtype EngineLog   = EngineLog (Seq String)
  deriving Monoid

data EngineState = EngineState { uniqueSupply    :: !UniqSupply
                               , experimentCache :: !(Maybe ExperimentCache)
                               , randomGen       :: !StdGen
                               }

data EngineEnv = EngineEnv { maxExpressionTime   :: !Int
                           , maxExpressionAllocs :: !Int64
                           }

mkLabels [''EngineState, ''EngineEnv]

instance GhcMonad EngineM where
  getSession = EngineM $ lift getSession
  setSession = EngineM . lift . setSession

instance MonadUtils.MonadIO EngineM where
  liftIO = EngineM . lift . liftIO

instance MonadUnique EngineM where
  getUniqueSupplyM = EngineM $ do
    (us1, us2) <- gets $ splitUniqSupply . uniqueSupply
    modify (\s -> s { uniqueSupply = us2 })
    return us1

instance ExceptionMonad EngineM where
  gcatch (EngineM c) ((_runEngineM .) -> f) = EngineM $ do
    r <- ask
    s <- get
    let f' = ($ s) . ($ r) . runRWST . f
    mapRWST (\i -> gcatch i f') c

  gblock (EngineM c) = EngineM $ mapRWST gblock c
  gunblock (EngineM c) = EngineM $ mapRWST gunblock c

instance WarnLogMonad EngineM where
  setWarnings messages = EngineM $ lift (setWarnings messages)
  getWarnings = EngineM $ lift getWarnings

instance MonadThings EngineM where
  lookupThing = EngineM . lift . lookupThing


mkInitialEngineState :: Maybe Strategy -> Ghc EngineState
mkInitialEngineState mCacheStrategy = do
  uniqSupply <- liftIO $ mkSplitUniqSupply 'T'
  return $ EngineState { uniqueSupply = uniqSupply 
                       , experimentCache = EC.newCache <$> mCacheStrategy
                       , randomGen = mkStdGen 0xbadbabee
                       }

tellLog :: String -> EngineM ()
tellLog = EngineM . tell . EngineLog . Seq.singleton

lookupExperimentResult :: Experiment -> EngineM (Maybe ExperimentResult)
lookupExperimentResult exp = EngineM $ 
  gets (EC.lookupExperimentResult exp <=< experimentCache)

checkExperimentCache :: Experiment -> EngineM (Maybe Id)
checkExperimentCache exp = EngineM $ 
  gets (EC.lookupExperimentId exp <=< experimentCache)

storeExperimentResult :: Experiment -> ExperimentResult -> EngineM ()
storeExperimentResult exp res = EngineM $
  modify (mod lExperimentCache (EC.storeExperimentResult exp res <$>))

storeExperimentHValueAndId :: Experiment -> Id -> EngineHValue -> EngineM (Maybe (Seq Experiment))
storeExperimentHValueAndId exp newId hval = do
  mExisting <- EngineM $ gets (EC.lookupHValueId hval <=< experimentCache) 
  case mExisting of
    Nothing -> do 
      EngineM $ modify (mod lExperimentCache (EC.storeEngineHValueAndId exp hval (Just newId) <$>))
      EngineM . lift . liftIO $ extendLinkEnv [(varName newId, toHValue hval)]
      return Nothing
    Just (es, oldId) -> do
      EngineM $ modify (mod lExperimentCache (EC.storeEngineHValueAndId exp hval Nothing <$>))
      return (Just es)

runEngineM :: EngineState -> EngineEnv -> EngineM a -> Ghc (a, EngineState, EngineLog)
runEngineM st env code = runRWST ( _runEngineM $ code) env st

nextRandomR :: Int -> Int -> EngineM Int
nextRandomR lo hi = EngineM $ do
  st <- get 
  let (a, g) = randomR (lo,hi) (randomGen st)
  put $ st { randomGen = g }
  return a
