{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE PatternGuards #-}
{-# LANGUAGE BangPatterns #-}
module IdDB.Impl
( IdDB
, functionsToTest
, mkIdDBForModule
, mkIdDBForIds
, rawIdsForType
, idsForType
, constructorsForTyCon
, IdSource(..)
) where

import Control.Applicative
import Control.Arrow
import Control.Category
import Control.Monad
import Control.Monad.State.Strict
import Control.Monad.Trans
import CoreSyn
import DataCon
import Data.Foldable (Foldable)
import Data.List
import Data.Maybe
import Data.Monoid
import GHC
import HscMain
import HscTypes
import Id
import IdDB.TypeMapDebug
import IdDB.TypeMapDebug
import IdDB.TypeMap (TypeMap)
import IdInfo
import Name
import Outputable hiding (trace)
import Prelude hiding (id, mod, (.))
import qualified Data.Foldable as F
import qualified IdDB.TypeMap as TM
import qualified Util.UniqMap as UniqMap
import qualified Util.UniqSet as UniqSet
import SCA.Simple
import TyCon
import Type
import TysWiredIn
import UniqSupply
import Util
import Util.FCLabels as FC
import Util.GHC
import Util.UniqMap ( UniqMap )
import Util.UniqSet ( UniqSet )
import Var
import VarEnv

data IdDB = IdDB { tyMap      :: !TypeMap 
                 , dataConMap :: !(UniqMap TyCon (UniqSet DataCon))
                 , testFns    :: ![Id] 
                 }

mkLabels [''IdDB]

functionsToTest :: IdDB -> [Id]
functionsToTest = testFns

instance Show IdDB where
  show IdDB { tyMap = tyMap
            , testFns = testFns
            } = "IdDB { tyMap = " ++ showPpr tyMap ++ ", testFns = " ++ show testFns ++ "}"

data IdSource = None
              | TestModule 
              | SourceModule 
              | PackageNearestVisibleModules 
              | PackageAllModules
  deriving (Eq, Show)

data ExtendIdDBSource = ESourceModule
                      | EPackageNearestVisibleModules
                      | EPackageAllModules
  deriving (Eq, Show)

foldSource :: IdSource -> a -> a -> (ExtendIdDBSource -> a) -> a
foldSource source none test others = case source of
  None                         -> none
  TestModule                   -> test
  SourceModule                 -> others ESourceModule
  PackageNearestVisibleModules -> others EPackageNearestVisibleModules
  PackageAllModules            -> others EPackageAllModules

emptyIdDB :: IdDB
emptyIdDB = IdDB { tyMap      = TM.empty 
                 , dataConMap = mempty
                 , testFns    = []
                 }

extendIdDB :: ExtendIdDBSource -> IdDB -> Id -> StateT (UniqSet TyCon) Ghc IdDB
extendIdDB idSource iddb id = do
  tm <- foldM (lookupAndMaybeAddProvidersFor idSource) (tyMap iddb) leadTys
  return (iddb { tyMap = tm })
  where
    (leadTys, resTy) = splitFunTys . dropForAlls . varType $ id


lookupAndMaybeAddProvidersFor :: ExtendIdDBSource -> TypeMap -> Type -> StateT (UniqSet TyCon) Ghc TypeMap
lookupAndMaybeAddProvidersFor idSource tm ty = do
  case splitTyConApp_maybe ty of
    Just (tcon, argTys) -> do
      seenBefore <- gets (UniqSet.member tcon)
      if seenBefore
        then return tm
        else do
          modify (`mappend` (UniqSet.singleton tcon))
          tm' <- lookupAndAddProvidersFor idSource tm tcon 
          foldM (lookupAndMaybeAddProvidersFor idSource) tm' argTys
    _ ->  {- TODO maybe something smarter here! -} return tm
    
lookupAndAddProvidersFor :: ExtendIdDBSource -> TypeMap -> TyCon -> StateT (UniqSet TyCon) Ghc TypeMap
lookupAndAddProvidersFor ESourceModule tm tcon 
  | isClassTyCon tcon = {- do
    modInfo <- fromMaybe (error $ "Couldn't lookup modinfo for: " ++ showPpr sourceMod) 
           <$> lift (getModuleInfo sourceMod)
    let possibleInstances = filter (((getName tcon) `elemNameSet`) 
                                   . tyClsNamesOfDFunHead 
                                   . idType 
                                   . instanceDFunId) 
                          . modInfoInstances $ modInfo
    
    {- trace ("Class TyCon: " ++ showPpr tcon) -} ARGH Type classes are complicated! -} (return tm)
  | isAlgTyCon tcon = {- trace ("Looking up: " ++ showPpr tcon) $-} do
    modInfo <- fromMaybe (error $ "Couldn't lookup modinfo for: " ++ showPpr sourceMod) 
           <$> lift (getModuleInfo sourceMod)
    let tconWiredIn = ((tcon `elem` wiredInTyCons) ||)
    let availExports = modInfoExports modInfo
    let availDataCons = filter (tconWiredIn . (`elem` availExports) . dataConName) <$> dataCons

    -- () <- trace ("Source mod: " ++ showPpr sourceMod) (return ())
    -- () <- trace ("Data cons: " ++ showPpr dataCons) (return ())
    -- () <- trace ("Avail exports: " ++ showPpr availExports) (return ())
    -- () <- trace ("Avail data cons: " ++ showPpr availDataCons) (return ())
    -- () <- trace ("TM:" ++ showPpr tm) (return ())

    if (isJust dataCons && availDataCons == dataCons)
      then do
        let ids = map dataConWorkId (fromJust availDataCons)
        {- TODO, records, case ctrs, etc -}
        {- () <- trace ("Adding: " ++ show ids) (return ()) -} 
        (return $ foldl' TM.addProvider tm ids)
      else do
        allAvailIds <- catMaybes <$> mapM ((getTyThingId <$>) . lift . lookupThing) availExports
        -- () <- trace ("Candidates: " ++ showPpr candidates) (return ())
        let candidates = do
                          availId <- allAvailIds
                          let (srcTypes, resTy) = splitFunTys . dropForAlls . varType $ availId
                          Just (tcon', _) <- return $ splitTyConApp_maybe resTy
                          guard (tcon' == tcon)
                          return (availId, srcTypes)
        let tm' = foldl' TM.addProvider tm (map fst candidates)
        -- () <- trace ("Candidates: " ++ showPpr candidates) (return ())
        foldM (lookupAndMaybeAddProvidersFor ESourceModule) tm' (concatMap snd candidates)
  | otherwise = {- trace ("Rejected: " ++ showPpr tcon) -} ( return tm) {- TODO maybe something smarter here! -}
  where
    sourceMod = nameModule . getName $ tcon
    dataCons = tyConDataCons_maybe tcon
lookupAndAddProvidersFor _ _ _ = error "TODO lookupAndAddProvidersFor non ESourceModule"
   
mkIdDBForModule :: Module -> IdSource -> [Module] -> Ghc IdDB
mkIdDBForModule mdl autoIdSource explicitIdSources = do
  testIds <- filter (isVanillaIdDetails . idDetails) <$> getAvailableIds mdl
  mkIdDBForIds testIds autoIdSource explicitIdSources

mkIdDBForIds :: [Id] -> IdSource -> [Module] -> Ghc IdDB
mkIdDBForIds ids source explicitSources = do 
  sourcedIdDB <- foldSource source (return $ emptyIdDB { testFns = ids }) 
                            testModuleSources 
                            $ \extendedSource -> do
                    evalStateT (foldM (extendIdDB extendedSource) 
                                      (emptyIdDB { testFns = ids }) ids) mempty
  providedIdDB <- addProviderModules explicitSources sourcedIdDB
  dataCons <- foldM addConstructorsForId mempty ids

  return $ providedIdDB { dataConMap = dataCons }
  where 
    modSets = F.foldMap (UniqSet.singleton . nameModule . idName) ids

    testModuleSources = do
      allIds <- F.foldlM (\ids mod -> (ids ++) <$> getAvailableIds mod) [] modSets
      -- () <- return $ trace ("All ids: " ++ show allIds) ()
      return $ emptyIdDB { tyMap   = foldl' TM.addProvider TM.empty allIds
                         , testFns = ids
                         }

addConstructorsForId :: (GhcMonad m) => UniqMap TyCon (UniqSet DataCon) -> Id -> m (UniqMap TyCon (UniqSet DataCon))
addConstructorsForId !inMap !id = addConstructorsForType inMap (idType id)
  where
    addConstructorsForType :: (GhcMonad m) => UniqMap TyCon (UniqSet DataCon) -> Type -> m (UniqMap TyCon (UniqSet DataCon))
    addConstructorsForType !inMap !ty 
      | Just (tc, _) <- mLastDataCon = addConstructorsForTyCon inMap tc 
      | otherwise                    = return inMap
      where
        mLastDataCon = splitTyConApp_maybe . last . listFunTys . snd . splitForAllTys $ ty

    addConstructorsForTyCon :: (GhcMonad m) => UniqMap TyCon (UniqSet DataCon) -> TyCon -> m (UniqMap TyCon (UniqSet DataCon))
    addConstructorsForTyCon !inMap !tc
      | tc `UniqMap.member` inMap = return inMap
      | isDataTyCon tc 
      , Just mod <- mModule = do
        let tconWiredIn = ((tc `elem` wiredInTyCons) ||)
        availExports <- (maybe [] modInfoExports) <$> getModuleInfo mod
        let possibleDCS = filter (tconWiredIn . (`elem` availExports) . dataConName) . tyConDataCons $ tc

        let lvl1OutMap = inMap `UniqMap.insert` tc $ F.foldMap UniqSet.singleton possibleDCS
        foldM (\inMap -> foldM addConstructorsForType inMap . dataConOrigArgTys) lvl1OutMap possibleDCS
      | otherwise = return inMap
      where
        mModule = nameModule_maybe . getName $ tc

rawIdsForType :: Type -> IdDB -> [(Id, TvSubst)]
rawIdsForType ty db = do
  (subst, ids) <- TM.lookupProviders ty (tyMap db) 
  id <- UniqSet.toList ids
  return (id, subst)


idsForType :: InScopeSet -> Type -> IdDB -> [(Id, TvSubst, InScopeSet)]
idsForType env ty iddb = do
  --() <- return $ trace ("idsForType: " ++ showDebug ty) ()
  (id, _) <- rawIdsForType ty iddb
  let (env', id', _) = alphaConvert env id

  subst <- case matchTypes (idType id') ty
    of Just s -> return s
       Nothing -> {- TODO type classes break this trace ("idsForType: " ++ showDebug ty ++ " thought I got: " ++ showDebug id ++ "( " ++ showDebug (idType id) ++ ")  which became: " ++ showDebug id' ++ "( " ++ showDebug (idType id') ++ ") but it didn't match!") -} []
  --() <- return $ trace (" Got: " ++ showDebug id') ()
  return $ (id', subst, env') 
  where
    matchTypes :: Type -> Type -> Maybe TvSubst
    matchTypes source target
      | numTargetComponents > numSourceComponents = Nothing
      | otherwise = t2s
      where
        (_, sourceRawType) = splitForAllTys source
        (_, targetRawType) = splitForAllTys target

        (sourceComponents, sourceLast) = splitFunTys sourceRawType

        numTargetComponents = length $ listFunTys targetRawType
        numSourceComponents = 1 + length sourceComponents

        sourceTail = mkFunTys (drop (numSourceComponents - numTargetComponents) sourceComponents) sourceLast

        t2s = unifyTypes targetRawType sourceTail


constructorsForTyCon :: TyCon -> IdDB -> UniqSet DataCon
constructorsForTyCon tc (dataConMap -> m) = UniqMap.lookup m tc

addProviderId :: IdDB -> Id -> IdDB
addProviderId iddb id = FC.mod lTyMap (flip TM.addProvider id) iddb 

addProviderIds :: [Id] -> IdDB -> IdDB
addProviderIds ids db = foldl' addProviderId db ids

addProviderModules :: [Module] -> IdDB -> Ghc IdDB
addProviderModules = flip $ foldM (flip addProviderModule)

addProviderModule :: Module -> IdDB -> Ghc IdDB
addProviderModule mod iddb = flip addProviderIds iddb <$> getAvailableIds mod


constantsForId :: Id -> Ghc Constants
constantsForId id = do
  mModGuts <- modGutsForId id
  return . maybe mempty findConstants $ 
    lookup id . flattenBinds . mg_binds =<< mModGuts

modGutsForId :: Id -> Ghc (Maybe ModGuts)
modGutsForId id = do
  let mMod = nameModule_maybe . idName $ id  
  case mMod of
    Nothing  -> return Nothing
    Just mod -> do
      modSummary <- getModSummary . moduleName $ mod
      Just <$> withTempSession ( Control.Category.id )
                                 -- set (lHpcDir . lHsc_dflags) ".hpc2") 
                                 -- error . show . FC.get (lOpt_F . lHsc_dflags)
                               (hscSimplify . coreModule =<< desugarModule 
                                                         =<< typecheckModule 
                                                         =<< parseModule modSummary)

