{-# LANGUAGE TypeOperators #-}
{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# OPTIONS_GHC -fno-warn-overlapping-patterns #-}
module Main where

{- With a lot of inspiration taken from cabal-install 
 - 2005 Lemmih <lemmih@gmail.com> 
 - 2006 Paolo Martini <paolo@nemail.it> 
 - 2007 Bjorn Bringert <bjorn@bringert.net> 
 - 2007 Isaac Potoczny-Jones <ijones@syntaxpolice.org> 
 - 2008-2009 Duncan Coutts <duncan@haskell.org>
 -}

import Data.Monoid
import Engine (Strategy, fifoStrategy)
import Distribution.Simple.Command hiding (boolOpt)
import Distribution.Simple.Setup
import Format
import FrontEnd.Options
import FrontEnd.Source
import IdDB
import System.Environment
import System.Exit              (exitFailure)
import Util.FCLabels

main :: IO ()
main = getArgs >>= mainWorker

mainWorker :: [String] -> IO ()
mainWorker args 
  = case commandsRun irulanCommand commands args of
      CommandHelp help    -> printHelp help
      CommandList opts    -> printOptionsList opts
      CommandErrors errs  -> printErrors errs
      CommandReadyToGo (options, commandParse) -> case commandParse of
        CommandHelp help         -> printHelp help
        CommandList opts         -> printOptionsList opts
        CommandErrors errs       -> printErrors errs
        CommandReadyToGo action  -> action options
      
  where
    commands = [sourceCommand `commandAddAction` sourceAction]

    printHelp help = do
      pname <- getProgName
      putStr (help pname)

    printOptionsList = putStr . unlines

    printErrors errs = do
      putStrLn "Print errors!"
      putStr . unlines $ errs
      exitFailure

irulanCommand :: CommandUI IrulanOptions
irulanCommand 
  = CommandUI { commandName         = ""
              , commandSynopsis     = ""
              , commandUsage        = const . unlines $ 
                  [ "This is irulan by Tristan Allwood."
                  , "http://www.doc.ic.ac.uk/~tora/irulan/"
                  ]
              , commandDescription  = Nothing 
              , commandDefaultFlags = defaultOptions
              , commandOptions      = const
                  [ optionL [] ["search-strategy"]
                      "Choose the type of search strategy"
                      lRuntimePlanType
                      (choiceOpt [ ( toFlag DepthBounded
                                   , (['d'],["depth-bounded"])
                                   , "A complete, depth-bounded strategy. Configure the bound size with --depth"
                                   )
                                 , ( toFlag RandomTree
                                   , (['r'], ["random-tree"])
                                   , "A random strategy.  Configure the maximum runtime per function-to-test with --functionRuntime"
                                   )
                                 ]
                      )
                  , optionL [] ["depth"]
                      "The maximum depth of the depth-bounded search strategy (if chosen)"
                      lMaximumDepth
                      (reqArg' "INT" (toFlag . fromInteger . read) 
                                     (fmap format . flagToList)
                      )
                  , optionL [] ["runtimePerFunction"]
                      "The maximum time per function being tested for the random-tree search strategy (if chosen)"
                      lMaxRuntimePerFunction
                      (reqArg' "SECONDS" (toFlag . read)
                                         (fmap show . flagToList)
                      )
                  , optionL [] ["expressionTime"]
                      "The maximum time before a test expression times out"
                      lMaxExpressionTime
                      (reqArg' "MICROSECONDS" (toFlag . read)
                                              (fmap show . flagToList)
                      )
                  , optionL [] ["expressionAllocs"]
                      "The maximum number of allocations a test expression can perform before it is killed"
                      lMaxExpressionAllocs
                      (reqArg' "BYTES" (toFlag . read)
                                       (fmap show . flagToList)
                      )
                  , optionL ['t'] ["trace"]
                      "Enable tracing of test expressions as they are executed"
                      lOutputExperimentTrace
                      trueArg
                  , optionL ['v'] ["debug"]
                      "Enable debugging information"
                      lOutputDebugInformation
                      trueArg
                  , optionL [] ["stats-output"]
                      "Enable output of code coverage and errors found for graphing after every execution step"
                      lOutputStatistics
                      (reqArg' "FILENAME" toFlag flagToList)
                  , optionL [] ["simple-output"]
                      "simple output of expressions"
                      lUseSimpleOutput
                      (boolOpt [] [])
                  , optionL [] ["case-statements"]
                      "case statements in the plan"
                      lUseCaseStatement
                      (boolOpt [] [])
                  , optionL [] ["automatic-identifier-source"]
                      "Choose the source of automatically discovered identifiers used to build test expressions"
                      lSourceIdsFrom
                      (choiceOpt [ ( toFlag None
                                   , ([], ["no-ai"])
                                   , "Don't automatically include identifiers to use for testing."
                                   )
                                 ,
                                   ( toFlag TestModule
                                   , ([], ["ai-test"])
                                   , "When an argument of type T is needed, automatically include any identifier from the module being tested."
                                   )
                                 , ( toFlag SourceModule
                                   , ([], ["ai-source"])
                                   , "When an argument of type T is needed, lookup the declaring module for T and use identifiers from that module to make T."
                                   )
                                 ]
                      )
                  , optionL ['i'] ["include"]
                      "Choose the source modules used to provide identifiers used for testing"
                      lExplicitIdSource
                      (reqArg' "MODULENAME" (:[]) id)
                  , optionL [] ["no-cache"]
                      "disable runtime experiment value cache"
                      lCache
                      (noArg (Last (Just Nothing)))
                  , optionL [] ["fifo-cache"]
                      "enable runtime experiment value cache with a fifo-queue"
                      lCache
                      (reqArg' "MAXSIZE" (Last . Just . Just . fifoStrategy . read)
                                         (const ["<maxsize>"])
                      )
                  , optionL [] ["quiet"]
                      "be quiet when running experiments - don't output a progress indicator"
                      lQuiet
                      (boolOpt [] [])
                  ]
              }

sourceCommand :: CommandUI [String]
sourceCommand 
  = CommandUI { commandName         = "source"
              , commandSynopsis     = "Compile (if necessary) and then test source files"
              , commandUsage        = const "Test the identifiers in the provided source files, compiling them if necessary"
              , commandDescription  = Nothing
              , commandDefaultFlags = []
              , commandOptions      = const 
                  [ option [] ["ghc-options"]
                      "Additional ghc options"
                      id (++)
                      (reqArg' "OPTIONS" (:[]) id)
                  ]
                                         
              } 

optionL :: SFlags -> LFlags -> Description -> (a :-> b) -> MkOptDescr (a -> b) (b -> a -> a) a1 -> OptionField a1
optionL a b d l o = option a b d (get l) (set l) o

{- urk nothing is ever simple is it? -}
instance Eq Strategy where
  _ == _ = False
