{-# LANGUAGE RankNTypes #-}
{-# LANGUAGE ExistentialQuantification #-}
{-# LANGUAGE TemplateHaskell #-}
{-# OPTIONS_GHC -fno-warn-overlapping-patterns #-}
{-# LANGUAGE PatternGuards #-}
{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE BangPatterns #-}
module SCA.Simple where

import Control.Applicative
import Control.Arrow
import CoreSyn
import CoreUtils
import Data.Foldable (Foldable)
import Data.Monoid
import Data.Set (Set)
import FastString
import GHC hiding (compileExpr, exprType)
import HscMain (compileExpr)
import HscTypes
import Id
import Linker (extendLinkEnv)
import Literal
import MkCore
import qualified Data.Foldable as F
import qualified Data.Set as Set
import qualified Util.UniqSet as UniqSet
import UniqSupply
import Unique
import Util
import Util.FCLabels
import Util.GHC ()
import Util.UniqSet (UniqSet)
import Var

data Constants 
  = Constants 
  { charConstants   :: !(Set Char) 
  , fastStringConstants :: !(UniqSet FastString)
  , nullPtr         :: !Bool
  , intConstants    :: !(Set Integer)
  , int64Constants  :: !(Set Integer)
  , wordConstants   :: !(Set Integer)
  , word64Constants :: !(Set Integer)
  , floatConstants  :: !(Set Rational)
  , doubleConstants :: !(Set Rational)
  }
  deriving (Eq)

mkLabels [''Constants]

instance Monoid Constants where
  
  mempty = Constants { charConstants   = mempty
                     , fastStringConstants = mempty
                     , nullPtr         = False
                     , intConstants    = mempty
                     , int64Constants  = mempty
                     , wordConstants   = mempty
                     , word64Constants = mempty
                     , floatConstants  = mempty
                     , doubleConstants = mempty
                     }
    
  c1 `mappend` c2 = Constants { charConstants   = charConstants   c1 `mappend` charConstants   c2 
                              , fastStringConstants = fastStringConstants c1 `mappend` fastStringConstants c2 
                              , nullPtr         = nullPtr         c1    ||     nullPtr         c2 
                              , intConstants    = intConstants    c1 `mappend` intConstants    c2 
                              , int64Constants  = int64Constants  c1 `mappend` int64Constants  c2 
                              , wordConstants   = wordConstants   c1 `mappend` wordConstants   c2 
                              , word64Constants = word64Constants c1 `mappend` word64Constants c2 
                              , floatConstants  = floatConstants  c1 `mappend` floatConstants  c2 
                              , doubleConstants = doubleConstants c1 `mappend` doubleConstants c2 
                              }

findBindConstants :: CoreBind -> Constants
findBindConstants (NonRec _ e) = findConstants e
findBindConstants (Rec es)     = F.foldMap (findConstants . snd) es

findConstants :: CoreExpr -> Constants
findConstants (Var _) = mempty
findConstants (Lit l) = findLitConstants l
findConstants (App e1 e2)       = findConstants e1 `mappend` findConstants e2
findConstants (Lam _ e)         = findConstants e
findConstants (Let b e)         = findBindConstants b `mappend` findConstants e
findConstants (Case e _ _ alts) = findConstants e `mappend` (F.foldMap findAltConstants alts)
findConstants (Cast e _)        = findConstants e
findConstants (Note _ e)        = findConstants e
findConstants (Type _)          = mempty
findConstants _                 = error "findConstants dropped off the end"

findLitConstants :: Literal -> Constants
findLitConstants l
  | MachChar c   <- l = mempty { charConstants = Set.singleton c }
  | MachStr fs   <- l = mempty { fastStringConstants = UniqSet.singleton fs }
  | MachInt i    <- l = mempty { intConstants = Set.singleton i }
  | MachInt64 i  <- l = mempty { int64Constants = Set.singleton i }
  | MachFloat r  <- l = mempty { floatConstants = Set.singleton r }
  | MachDouble r <- l = mempty { doubleConstants = Set.singleton r }
  | MachWord i   <- l = mempty { wordConstants = Set.singleton i }
  | MachWord64 i <- l = mempty { word64Constants = Set.singleton i }

findAltConstants :: CoreAlt -> Constants
findAltConstants (con, _, e) = findConstants e `mappend` case con of
                                LitAlt l -> findLitConstants l
                                _        -> mempty
                                


data ConstantConverter 
  = forall f a . (Foldable f) => P  (a -> String) (Constants -> f a) (a -> CoreExpr)
  | forall f a . (Foldable f) => PM (a -> String) (Constants -> f a) (MonadThings m => a -> m CoreExpr)


constantsToIds :: UniqSupply -> Constants -> Ghc [Id]
constantsToIds us cs = do
  esAndNs <- (unzip . concat) <$> mapM (flip mkExprsAndNames cs) translations
  let uniqS = uniqsFromSupply us
  uncurry (zipWithM3 compileConstantExpr uniqS) esAndNs

translations :: [ConstantConverter]
translations = [ P  show         charConstants       mkCharExpr
               , P  show         intConstants        mkIntExpr
               , PM show         int64Constants      mkIntegerExpr
               , P  show         wordConstants       mkWordExpr
               , P  show         word64Constants     mkWordExpr
               , P  showAsFloat  floatConstants     (mkFloatExpr  . fromRational)
               , P  showAsDouble doubleConstants    (mkDoubleExpr . fromRational)
               , PM showAsString fastStringConstants mkStringExprFS
               -- TODO make a null ptr , P  (Just . nullPtr) (
               ]
  where
    showAsFloat :: Rational -> String
    showAsFloat = (show :: Float -> String) . fromRational

    showAsDouble :: Rational -> String
    showAsDouble = (show :: Double -> String) . fromRational

    showAsString :: FastString -> String
    showAsString = show . unpackFS


mkExprsAndNames :: (MonadThings m) => ConstantConverter -> Constants -> m [(String, CoreExpr)]
mkExprsAndNames (P   asString extract build) 
  = return . map (asString &&& build) . F.toList . extract
mkExprsAndNames (PM  asString extract build) 
  = mapM (runKleisli $ arr asString &&& Kleisli build) . F.toList . extract

compileConstantExpr :: Unique -> String -> CoreExpr -> Ghc Id
compileConstantExpr unique name expr = do
  let letId = mkSysLocal (mkFastString name) unique (exprType expr)
  let codeToRun = mkCoreLet (NonRec letId expr) (Var letId)
  hval <- withSession $ \hsc -> liftIO $ compileExpr hsc noSrcSpan codeToRun 
  liftIO $ extendLinkEnv [(varName letId, hval)]
  return letId
